import 'dart:async';

import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:flutter/material.dart';
import 'package:flutter_animate/flutter_animate.dart';
import 'package:healthrian_visualization_support/healthrian_visualization_support.dart';
import 'package:rxdart/rxdart.dart';

part 'ble_scanner.g.dart';

_$BleScannerCWProxy get ScannerWearEcg =>
    BleScanner(module: BLEManager().wearEcg12.module).copyWith;

_$BleScannerCWProxy get ScannerCardioClick =>
    BleScanner(module: BLEManager().cardioClick.module).copyWith;

_$BleScannerCWProxy get ScannerWearT =>
    BleScanner(module: BLEManager().wearCardioT.module).copyWith;

@CopyWith()
class BleScanner extends StatefulWidget {
  const BleScanner({
    super.key,
    this.module,
    this.connectedBuilder,
    this.scannedBuilder,
    this.resetButton,
    this.emptyWidget,
    this.onTapConnected,
    this.onTapScanned,
    this.scanFrequency = 3,
    this.enableTapConnected = true,
    this.enableTapScanned = true,
    this.enableResetButton = false,
  });

  final ModuleInterface? module;

  final Function(BluetoothDevice)? onTapConnected;
  final Function(ScanResult)? onTapScanned;
  final Widget Function(BluetoothDevice)? connectedBuilder;
  final Widget Function(ScanResult)? scannedBuilder;
  final Widget? resetButton;
  final Widget? emptyWidget;

  final int scanFrequency;

  final bool enableTapConnected;
  final bool enableTapScanned;
  final bool enableResetButton;

  @override
  State<BleScanner> createState() => _BleScannerState();
}

class _BleScannerState extends State<BleScanner> {
  late final module = widget.module ?? BLEManager().scanner;

  final _controller = StreamController<Map<DeviceIdentifier, ScanResult>>();
  late final _streamScan = _controller.stream.asBroadcastStream();
  final _scanMap = <DeviceIdentifier, ScanResult>{};

  late final Stream<List<BluetoothDevice>> _streamConnect;
  late final Stream<MergeScanConnect> _streamMerge;

  @override
  void initState() {
    super.initState();
    _streamConnect = Rx.merge([
      Stream.value(FlutterBluePlus.connectedDevices),
      Stream.periodic(500.ms, (_) => FlutterBluePlus.connectedDevices),
    ]).distinct();

    _streamMerge = _streamScan
        .zipWith(_streamConnect, (t, s) => (t, s))
        .asBroadcastStream();

    _rescan();
  }

  @override
  void dispose() {
    super.dispose();
    _controller.close();
  }

  Future<void> _rescan() async {
    if (FlutterBluePlus.isScanningNow) return;
    final subscription = FlutterBluePlus.scanResults.listen(
      (results) {
        for (ScanResult r in results) {
          if (_scanMap[r.device.remoteId]?.device.platformName.isEmpty == false)
            continue;
          _scanMap[r.device.remoteId] = r;
          _controller.add(_scanMap);
        }
      },
    );

    final timeout = const Duration(seconds: 4);
    await FlutterBluePlus.startScan();
    await Future.delayed(timeout);
    await FlutterBluePlus.stopScan();
    subscription.cancel();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      backgroundColor: Colors.transparent,
      body: StreamBuilder(
        stream: _streamMerge,
        builder: (context, snapshot) {
          final scanned = [
            ...?snapshot.data?.$1.values
                .where((sr) => module.ble.scanFilter(sr.device))
          ]..sort((a, b) => b.rssi - a.rssi);
          final connected = [
            ...?snapshot.data?.$2.where(module.ble.scanFilter)
          ];
          if (scanned.length + connected.length == 0) {
            return widget.emptyWidget ?? SizedBox();
          }
          return ListView.builder(
            itemCount: connected.length + scanned.length,
            itemBuilder: (context, index) {
              /// connected
              if (index < connected.length) {
                final device = connected[index];
                return InkWell(
                  onTap: widget.enableTapConnected
                      ? () => _onTapConnected(module, device)
                      : null,
                  child: _getChildConnected(device, context),
                );
              }

              /// scanned
              final mIndex = index - connected.length;
              final scanResult = scanned[mIndex];
              return InkWell(
                onTap: widget.enableTapScanned
                    ? () => _onTapScanned(module, scanResult)
                    : null,
                child: _getChildScanned(scanResult, context),
              );
            },
          );
        },
      ),
      floatingActionButton: widget.enableResetButton ? _resetButton() : null,
    );
  }

  Widget _resetButton() {
    return FloatingActionButton(
      onPressed: () {
        _scanMap.clear();
        _rescan();
      },
      child: widget.resetButton ?? Icon(Icons.refresh),
    );
  }

  Future<void> _onTapScanned(
      ModuleInterface module, ScanResult scanResult) async {
    await module.ble
        .connectDevice(targetDevice: scanResult.device.remoteId.str);
    _scanMap.remove(scanResult.device.remoteId);
    _controller.add(_scanMap);
    widget.onTapScanned?.call(scanResult);
  }

  Widget _getChildScanned(ScanResult scanResult, BuildContext context) {
    return widget.scannedBuilder?.call(scanResult) ??
        Card(
          child: SizedBox(
            height: 80,
            child: ListTile(
              title: Text('${scanResult.device.platformName}'),
              titleTextStyle: Theme.of(context)
                  .textTheme
                  .titleLarge
                  ?.copyWith(fontSize: 24),
              subtitle: Text(scanResult.device.remoteId.str),
              subtitleTextStyle: Theme.of(context)
                  .textTheme
                  .titleLarge
                  ?.copyWith(fontSize: 16),
              trailing: Text(
                'Scanned',
                style: Theme.of(context).textTheme.titleSmall,
              ),
            ),
          ),
        );
  }

  Future<void> _onTapConnected(
      ModuleInterface module, BluetoothDevice device) async {
    await module.ble.disconnectDevice(device: device);
    _rescan();
    widget.onTapConnected?.call(device);
  }

  Widget _getChildConnected(BluetoothDevice device, BuildContext context) {
    return widget.connectedBuilder?.call(device) ??
        Card(
          child: SizedBox(
            height: 80,
            child: ListTile(
              title: Text('${device.platformName}'),
              titleTextStyle: Theme.of(context)
                  .textTheme
                  .titleLarge
                  ?.copyWith(fontSize: 24),
              subtitle: Text(device.remoteId.str),
              subtitleTextStyle: Theme.of(context)
                  .textTheme
                  .titleLarge
                  ?.copyWith(fontSize: 16),
              trailing: Text(
                'Connected',
                style: Theme.of(context).textTheme.titleSmall,
              ),
            ),
          ),
        );
  }
}
